/*
 *               ParaStation
 *
 * Copyright (C) 2003-2004 ParTec AG, Karlsruhe
 * Copyright (C) 2005 Cluster Competence Center GmbH, Munich
 *
 * $Id: adminparser.h 3880 2006-01-02 10:17:35Z eicker $
 *
 */
/**
 * \file
 * ParaStation admin command line parser functions
 *
 * $Id: adminparser.h 3880 2006-01-02 10:17:35Z eicker $
 *
 * \author
 * Norbert Eicker <eicker@par-tec.com>
 *
 */
#ifndef __ADMINPARSER_H
#define __ADMINPARSER_H

#ifdef __cplusplus
extern "C" {
#if 0
} /* <- just for emacs indentation */
#endif
#endif

/**
 * @brief Parse admins input line.
 *
 * Parse a single line conforming to the syntax of psiadmin an execute
 * to corresponding commands.
 *
 * @param line The line to handle, i.e. to parse and execute.
 *
 * @return If the 'exit' or 'quit' command was reached, 1 is returned
 * or 0 otherwise.
 */
int parseLine(char *line);

/**
 * @brief Generate candidate for line completion
 *
 * Generate a list list of candidates for readline's line completion
 * given the word @a text typed in so far. This function is intended
 * to act as readline's @a rl_attempted_completion_function, thus it
 * will be called from readline whenever a completion has to be done.
 *
 * The actual implementation utilizes the same keylists as the parser
 * used to implement the special syntax of psiadmin's directives.
 *
 * @param text The text to complete
 *
 * @param start The position of the first character of @a text within
 * readline's @a rl_line_buffer.
 *
 * @param end The position of the last character of @a text within
 * readline's @a rl_line_buffer.
 *
 * @return If possible completions are found, a array of these strings
 * is returned. The actual array returned is created using readline's
 * @ref rl_completion_matches() function.
 *
 * @see rl_line_buffer, rl_completion_matches()
 */
char **completeLine(const char *text, int start, int end);

#ifdef __cplusplus
}/* extern "C" */
#endif

#endif /* __ADMINPARSER_H */
